/*
 * kdumpfix: a small tool to tweak kdump cores, so they can be used with gdb
 *
 * Copyright (C) Jean-Marc Saffroy <saffroy@gmail.com> 2006
 * This program is free software, distributed under the terms of the
 * GNU General Public License version 2.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <libelf.h>
#include <errno.h>
#include <string.h>
#include <elf.h>
#include <sys/procfs.h>

#ifndef PAGE_SIZE
#define PAGE_SIZE (1 << 12)
#endif

#define warn(fmt,arg...) \
	do {							\
		fprintf(stderr, "warning: " fmt, ##arg);	\
	} while(0)

#define err(fmt,arg...) \
	do {						\
		fprintf(stderr, "error: " fmt, ##arg);	\
		exit(1);				\
	} while(0)

void usage(char *progname) {
	fprintf(stderr, "usage: %s <core file> <new file> [<remap file>] [<thread data>]\n",
		progname);
	exit(1);
}

/* array of program headers being made */
int num_phdr = 0, alloc_phdr = 0;
Elf64_Phdr *phdr_list = NULL;

void add_phdr(Elf64_Phdr *phdr) {
	if(num_phdr == alloc_phdr) {
		alloc_phdr += 10;
		phdr_list = realloc(phdr_list, alloc_phdr*sizeof(Elf64_Phdr));
		if(phdr_list == NULL)
			err("realloc of %ld bytes failed\n",
				alloc_phdr*sizeof(Elf64_Phdr));
	}
	phdr_list[num_phdr++] = *phdr;
}

/* create dummy section mapping the program headers matching given type */
void build_sect(Elf *elfin, Elf *elfout, Elf64_Word type) {
	Elf64_Ehdr *ehdrin;
	Elf64_Phdr *phdrin;
	char *raw;

	Elf_Scn *scn;
	Elf64_Shdr *shdr;
	Elf64_Off offset;
	int i;

	ehdrin = elf64_getehdr(elfin);
	phdrin = elf64_getphdr(elfin);
	if(!phdrin)
		err("elf_getphdr: %s\n", elf_errmsg(elf_errno()));
	raw = elf_rawfile(elfin, NULL);
	if(raw == NULL)
		err("elf_rawfile: %s\n", elf_errmsg(elf_errno()));

	/* create dummy data section */
	scn = elf_newscn(elfout);
	if(scn == NULL)
		err("elf_newscn: %s\n", elf_errmsg(elf_errno()));
	shdr = elf64_getshdr(scn);
	if(shdr == NULL)
		err("elf64_getshdr: %s\n", elf_errmsg(elf_errno()));
	shdr->sh_type = SHT_NOTE;
	shdr->sh_flags = SHF_ALLOC;
	shdr->sh_name = 0; // SHN_UNDEF;
	if(type != PT_NOTE)
		shdr->sh_addralign = PAGE_SIZE;

	/* record program headers, populate dummy data section */
	offset = 0;
	for(i = 0; i < ehdrin->e_phnum; i++) {
		Elf64_Word p_type = phdrin[i].p_type;
		Elf64_Phdr phdr;
		Elf_Data *data;

		if((p_type != PT_NOTE) && (p_type != PT_LOAD))
			err("unexpected program header type: %d\n", p_type);
		if(p_type != type)
			continue;
#if 0
		printf("% 2d: 0x%lx -> 0x%lx (0x%lx)\n",
			i, phdrin[i].p_offset, phdrin[i].p_vaddr,
			phdrin[i].p_memsz);
#endif
		phdr = phdrin[i];
		/* for now, p_offset is relative to dummy section */
		phdr.p_offset = offset; 
		add_phdr(&phdr);

		/* x86-64: fix some strangely mapped areas */
		if(phdr.p_vaddr >= 0xffffffff80000000) {
			phdr.p_vaddr &= 0xffff81007fffffff;
#if 0
			printf("% 2d: ... -> 0x%lx\n", i, phdr.p_vaddr);
#endif
			add_phdr(&phdr);
		}

		data = elf_newdata(scn);
		if(data == NULL)
			err("elf_newdata: %s\n", elf_errmsg(elf_errno()));
		data->d_size = phdrin[i].p_filesz;
		data->d_buf = raw + phdrin[i].p_offset;
		data->d_type = ELF_T_BYTE;

		offset += phdrin[i].p_filesz;
	}

}

/* fix phdr file offsets after libelf has computed the file layout */
void fix_phdr(Elf *elfout, Elf64_Word type, Elf64_Word sect) {
	Elf_Scn *scn;
	Elf64_Shdr *shdr;
	Elf64_Phdr *phdrout;
	int i;

	scn = elf_getscn(elfout, sect);
	if(scn == NULL)
		err("elf_getscn: %s\n", elf_errmsg(elf_errno()));
	shdr = elf64_getshdr(scn);
	if(shdr == NULL)
		err("elf64_getshdr: %s\n", elf_errmsg(elf_errno()));
	phdrout = elf64_getphdr(elfout);
	if(!phdrout)
		err("elf_getphdr: %s\n", elf_errmsg(elf_errno()));
	for(i = 0; i < num_phdr; i++)
		if(phdrout[i].p_type == type)
			phdrout[i].p_offset += shdr->sh_offset;
}

/* add a phdr remapping a vmalloc page to an identity-mapped address */
void add_remap(long vmap, long idmap)
{
	Elf64_Phdr phdr;
	int i;

	for(i = 0; i < num_phdr; i++) {
		if(phdr_list[i].p_type == PT_LOAD)
			break;
	}
	if(i == num_phdr)
		err("internal error at %s:%s:%d\n",
			__FILE__, __FUNCTION__, __LINE__);
	phdr = phdr_list[i];
	phdr.p_vaddr = vmap;
	phdr.p_memsz = PAGE_SIZE;
	phdr.p_filesz = PAGE_SIZE;
	for(i = 0; i < num_phdr; i++) {
		if((idmap >= phdr_list[i].p_vaddr)
			&& (idmap + PAGE_SIZE)
			<= (phdr_list[i].p_vaddr + phdr_list[i].p_memsz))
			break;
	}
	if(i == num_phdr)
		err("add_remap: could not find segment mapping %lx\n", idmap);
	phdr.p_offset = idmap - phdr_list[i].p_vaddr + phdr_list[i].p_offset;
	add_phdr(&phdr);
}

/* parse remaps file */
void parse_remaps(char *file)
{
	FILE *f;
	long vmap, idmap;
	char buf[100];

	f = fopen(file, "r");
	if(f == NULL)
		err("fopen(%s, \"r\"): %s\n", file, strerror(errno));
	for(;;) {
		switch( fscanf(f, "%lx -> %lx\n", &vmap, &idmap) ) {
		case 2:
			add_remap(vmap, idmap);
			break;
		case EOF:
			goto out;
		default:
			/* ignore line */
			fgets(buf, sizeof(buf), f);
			break;
		}
	}
out:
	fclose(f);
}

/* array of ELF notes being made */
#define ALIGN4(x) (((x)+3)&~3)
#define ELFCORENOTESIZE				\
 (ALIGN4(sizeof(Elf64_Nhdr))		\
  + ALIGN4(strlen("CORE")+1)			\
  + ALIGN4(sizeof(struct elf_prstatus)))

int num_notes = 0, alloc_notes = 0;
char *note_list = NULL;

void add_note(char *note) {
	if(num_notes == alloc_notes) {
		alloc_notes += 10;
		note_list = realloc(note_list, alloc_notes*ELFCORENOTESIZE);
		if(note_list == NULL)
			err("realloc of %ld bytes failed\n",
				alloc_notes*ELFCORENOTESIZE);
	}
	memcpy(note_list+num_notes*ELFCORENOTESIZE, note, ELFCORENOTESIZE);
	num_notes++;
}

/* add an ELF note describing given thread */
void add_thr(long ip, long sp, long bp, long pid)
{
	Elf64_Nhdr note;
	struct elf_prstatus status;
	char buf[ELFCORENOTESIZE];
	char *ptr;

	note.n_namesz = strlen("CORE") + 1;
	note.n_descsz = sizeof(struct elf_prstatus);
	note.n_type = NT_PRSTATUS;
	ptr = buf;
	memcpy(ptr, &note, sizeof(note));
	ptr += ALIGN4(sizeof(note));
	memcpy(ptr, "CORE", strlen("CORE")+1);
	ptr += ALIGN4(strlen("CORE")+1);
	memset(&status, 0, sizeof(status));
	status.pr_pid = pid;
	/* see ELF_CORE_COPY_REGS for array indexes */
	/* see arch_unw_init_blocked for principle */
	status.pr_reg[16] = ip;
	status.pr_reg[19] = sp;
	status.pr_reg[ 4] = bp;
	memcpy(ptr, &status, sizeof(status));
	add_note(buf);
}

/* parse thrdata file */
void parse_thrdata(char *file)
{
	FILE *f;
	long ip, sp, bp, pid;
	char buf[100];
	char comm[100];

	f = fopen(file, "r");
	if(f == NULL)
		err("fopen(%s, \"r\"): %s\n", file, strerror(errno));
	for(;;) {
		switch( fscanf(f, "thread_return %lx\n", &ip) ) {
		case 1:
			/* ip parsed */
			goto next;
		case EOF:
			goto out;
		default:
			/* ignore line */
			fgets(buf, sizeof(buf), f);
			break;
		}
	}
next:
	for(;;) {
		switch( fscanf(f, "%lx %lx %6ld %s\n", &sp, &bp, &pid, comm) ) {
		case 4:
			add_thr(ip, sp, bp, pid);
			break;
		case EOF:
			goto out;
		default:
			/* ignore line */
			fgets(buf, sizeof(buf), f);
			break;
		}
	}
out:
	fclose(f);
}

/* append data to notes section and fix corresponding phdr accordingly */
void fix_notes(Elf *elfout, Elf64_Word sect)
{
	Elf_Scn *scn;
	int i;
	Elf_Data *data;

	scn = elf_getscn(elfout, sect);
	if(scn == NULL)
		err("elf_getscn: %s\n", elf_errmsg(elf_errno()));

	data = elf_newdata(scn);
	if(data == NULL)
		err("elf_newdata: %s\n", elf_errmsg(elf_errno()));
	data->d_size = num_notes * ELFCORENOTESIZE;
	data->d_buf = note_list;
	data->d_type = ELF_T_BYTE;

	for(i = 0; i < num_phdr; i++) {
		if(phdr_list[i].p_type == PT_NOTE) {
			phdr_list[i].p_memsz += data->d_size;
			phdr_list[i].p_filesz += data->d_size;
			break;
		}
	}
}

int main(int argc, char** argv) {
	int fdin, fdout;
	Elf *elfin, *elfout;
	Elf64_Ehdr *ehdrin, *ehdrout;
	Elf64_Phdr *phdrout;
	char *remaps = NULL;
	char *thrdata = NULL;

	if(argc != 3 && argc != 4 && argc != 5)
		usage(argv[0]);
	if(argc >= 4)
		remaps = argv[3];
	if(argc >= 5)
		thrdata = argv[4];

	if(elf_version(EV_CURRENT) == EV_NONE)
		err("elf_version: libelf version mismatch\n");

	/* input core file */
	fdin = open(argv[1], O_RDONLY);
	if(fdin < 0)
		err("open(%s, O_RDONLY): %s\n", argv[1], strerror(errno));
	
	elfin = elf_begin(fdin, ELF_C_READ, NULL);
	if(!elfin)
		err("elf_begin: %s\n", elf_errmsg(elf_errno()));
	ehdrin = elf64_getehdr(elfin);
	if(!ehdrin)
		err("elf_getehdr: %s\n", elf_errmsg(elf_errno()));
	if(ehdrin->e_type != ET_CORE)
		err("input file is not a core file!\n");

	/* output core file */
	fdout = open(argv[2], O_RDWR|O_CREAT|O_TRUNC, 0600);
	if(fdout < 0)
		err("open(%s, O_RDWR): %s\n", argv[2], strerror(errno));
	
	elfout = elf_begin(fdout, ELF_C_WRITE, NULL);
	if(!elfout)
		err("elf_begin: %s\n", elf_errmsg(elf_errno()));
	ehdrout = elf64_newehdr(elfout);
	if(!ehdrout)
		err("elf_newehdr: %s\n", elf_errmsg(elf_errno()));
	*ehdrout = *ehdrin;

	/* create dummy sections from program headers */
	build_sect(elfin, elfout, PT_NOTE); // 1
	build_sect(elfin, elfout, PT_LOAD); // 2
	if(remaps != NULL)
		parse_remaps(remaps);
	if(thrdata != NULL)
		parse_thrdata(thrdata);
	fix_notes(elfout, 1);

	phdrout = elf64_newphdr(elfout, num_phdr);
	if(!phdrout)
		err("elf_newphdr: %s\n", elf_errmsg(elf_errno()));
	memcpy(phdrout, phdr_list, sizeof(Elf64_Phdr)*num_phdr);

	/* compute layout and fix program headers */
	if(elf_update(elfout, ELF_C_NULL) < 0)
		err("elf_update: %s\n", elf_errmsg(elf_errno()));
	elf_flagelf(elfout, ELF_C_SET, ELF_F_DIRTY|ELF_F_LAYOUT);
	fix_phdr(elfout, PT_NOTE, 1);
	fix_phdr(elfout, PT_LOAD, 2);

	/* write out with current layout */
	if(elf_update(elfout, ELF_C_WRITE) < 0)
		err("elf_update: %s\n", elf_errmsg(elf_errno()));

	if(elf_end(elfout) < 0)
		err("elf_end: %s\n", elf_errmsg(elf_errno()));
	close(fdout);
	if(elf_end(elfin) < 0)
		err("elf_end: %s\n", elf_errmsg(elf_errno()));
	close(fdin);

	return 0;
}

